/*
 Controls 433 MHz transmitter on Raspberry Pi
 Created: 11 dec. 2012
 Changed: 23 sept 2013: No ON switching at midnight (test)
 Author: Jan Breemer
 */

#define VersionString  "433 MHz Controller on Raspberry Pi, V 1.3 2013-09-23"

// better if services do not constantly write to standard out
#define logging 0

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>

// For GPIO setup. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define BCM2708_PERI_BASE  0x20000000
#define GPIO_BASE          (BCM2708_PERI_BASE + 0x200000) /* GPIO controller */
#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)
int  mem_fd;
char *gpio_mem, *gpio_map;
char *spi0_mem, *spi0_map;
volatile unsigned *gpio;
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))
#define GPIO_SET *(gpio+7)  // sets   bits which are 1. Ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1. Ignores bits which are 0

// Globals
#define MaxLines 32
#define BufferSize MaxLines*17
struct Channel {int group, channel, timeon, timeoff, times2transmit;} Channels[MaxLines+1];
char buffer[BufferSize];
char Tmp[4];


void error(const char *msg)
{
    perror(msg);
    exit(1);
} // error

void Send433 (int Group, int Key, int On, int *NtoGo)
// Group and Key range from 1 to 4, On is zero (off) or > zero (on)
// Group and Key numbers are interpreted modulo 4
// GPIO 7..10 are Group 1..4
// GPIO 21..24 are Key 1..4
// GPIO 17 is On
// GPIO 18 is Off
{
	long Mask;
	if (logging) printf ("Setting%3d%3d%3d Still to do: %d\n", Group, Key, On, *NtoGo-1 ); // diagnose
	Mask = (0x0F<<7) | (0x0F<<21) | (0x03<<17);
	*(gpio+10) = Mask; // clear all bits in Mask
	if (On)
		*(gpio+7) = (1<<(7+((Group-1)&0x03))) | (1<<(21+((Key-1)&0x03))) | (1<<17);
	else
		*(gpio+7) = (1<<(7+((Group-1)&0x03))) | (1<<(21+((Key-1)&0x03))) | (1<<18);
	sleep (1); // transmit the code for 1 second
	*(gpio+10) = Mask; // clear all the bits
	sleep (1); // be silent for another second
	(*NtoGo)--;
} // Send433


int MinutesSinceMidnight (void)
{
	time_t Now;
	struct tm block;

	Now = time (NULL);
	block = *localtime(&Now);
	return block.tm_hour*60 + block.tm_min;
} // MinutesSinceMidnight


void PrintTime()
{
	int  min, hr;

	min = MinutesSinceMidnight() % 60;
	hr = MinutesSinceMidnight() / 60;
	if (logging) printf ("%s%02d%s%02d\n\n", "Time is now ", hr, ":", min);
} // PrintTime


void ShowData (void)
{
	int i;

	for (i = 1; i <= MaxLines; i++)
		if (logging) printf  ("%3d%3d%5d%5d%5d\n",Channels[i].group, Channels[i].channel, Channels[i].timeon, Channels[i].timeoff, Channels[i].times2transmit);
	PrintTime();
} // ShowData


int WriteFile (void)
{
	FILE *fp;
	int i;

	fp = fopen ("settings.txt","w");
	if (fp == NULL) return -1;        // on error
	for (i = 1; i <= MaxLines; i++)
	{
   		fprintf (fp, "%6d%6d%6d%6d\n", Channels[i].group, Channels[i].channel, Channels[i].timeon, Channels[i].timeoff);
	}
	fclose (fp);
	return 0;    // ok
} // WriteFile


int ReadFile (void)
{
	FILE *fp;
	int i;

	for (i = 1; i <= MaxLines; i++)  // Clear the list
 	{
    Channels[i].group = 0;
	 	Channels[i].channel = 0;
		Channels[i].timeon = 0;
		Channels[i].timeoff = 0;
	}
	fp = fopen ("settings.txt","r");
	if (fp == NULL) return -1;
	for (i = 1; i <= MaxLines; i++)
	{
		fscanf (fp, "%d%d%d%d", &Channels[i].group, &Channels[i].channel, &Channels[i].timeon, &Channels[i].timeoff);
		Channels[i].times2transmit = 10;
	}
	fclose (fp);
	return 0;    // ok
}// ReadFile


void ProcessData (void)
{
	int Now, OldNow = 1, i;
	static int n;

	if ((n < 0) || (n > MaxLines)) n = 1; // safe start
	Now = MinutesSinceMidnight();
	if (Now != OldNow)
	{
		OldNow = Now;
		PrintTime();
		ShowData();
	}
	if (Now == 0)
		for (i = 1; i <= MaxLines; i++)
			Channels[i].times2transmit = 10;  // restore at midnight
	if (Channels[n].times2transmit > 0)
		{
		if (Channels[n].timeon == 9999)  // always on
			Send433 (Channels[n].group, Channels[n].channel, 1,&Channels[n].times2transmit);
		if (Channels[n].timeoff == 9999) // always off
			Send433 (Channels[n].group, Channels[n].channel, 0,&Channels[n].times2transmit);
			if ((Channels[n].timeon == Now) && (Now != 0))  // no ON switching atexactly midnight
				Send433 (Channels[n].group, Channels[n].channel, 1,&Channels[n].times2transmit);
			if ((Channels[n].timeoff == Now) && (Now != 1439)) // no OFF switching at exactly midnight
				Send433 (Channels[n].group, Channels[n].channel, 0,&Channels[n].times2transmit);
		}
	if (++n > MaxLines) n = 0;
} // ProcessData


void DecomposeMessage (void)
{
	int i, n;
	for (n = 1; n <= MaxLines; n++)
	{
		i = (n-1) * 17;
		Tmp[0] = buffer[i];
		Tmp[1] = buffer[i+1];
		Tmp[2] = buffer[i+2];
		Tmp[3] = buffer[i+3];
		Channels[n].group = atoi(Tmp);
		Tmp[0] = buffer[i+4];
		Tmp[1] = buffer[i+5];
		Tmp[2] = buffer[i+6];
		Tmp[3] = buffer[i+7];
		Channels[n].channel = atoi(Tmp);
		Tmp[0] = buffer[i+8];
		Tmp[1] = buffer[i+9];
		Tmp[2] = buffer[i+10];
		Tmp[3] = buffer[i+11];
		Channels[n].timeon = atoi(Tmp);
		Tmp[0] = buffer[i+12];
		Tmp[1] = buffer[i+13];
		Tmp[2] = buffer[i+14];
		Tmp[3] = buffer[i+15];
		Channels[n].timeoff = atoi(Tmp);
		Channels[n].times2transmit = 10;
	}
ShowData();
WriteFile();
}  // DecomposeMessage


void SetupTCP(void)
{
	int sockfd, newsockfd, portno, n, flags;
	socklen_t clilen;
	struct sockaddr_in serv_addr, cli_addr;

	portno = 4712;
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) error("ERROR opening socket");
	flags = fcntl(sockfd, F_GETFL, 0 );
	fcntl( sockfd, F_SETFL, flags | O_NONBLOCK );
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = INADDR_ANY;
	serv_addr.sin_port = htons(portno);
	if (bind(sockfd,
		 (struct sockaddr *) &serv_addr,
					sizeof(serv_addr)) < 0)
							error("ERROR on binding");
	listen(sockfd,5);

	for (;;)
		{
		clilen = sizeof(cli_addr);
		newsockfd = accept(sockfd,
				 (struct sockaddr *) &cli_addr,
				 &clilen);
		if (newsockfd >= 0)
		{
			bzero(buffer, BufferSize);
			n = read(newsockfd, buffer, BufferSize);
			if (n < 0) error("ERROR reading from socket");
			if (logging) printf("Here is the message: %s\n", buffer);
			DecomposeMessage();
			time_t Now;
			struct tm block;
			Now = time (NULL);
			block = *localtime(&Now);
			n = write (newsockfd, asctime(&block), 19);
			if (n < 0) printf("ERROR writing to socket\n");
		}
		ProcessData();
	}
	close(newsockfd);
	close(sockfd);
} // SetupTCP


void Setup_GPIO()
{
	/* open /dev/mem */
	if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0)
	{
		printf("can't open /dev/mem \n");
		exit (-1);
	}
	// Allocate MAP block
	if ((gpio_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1))) == NULL)
	{
		printf("allocation error \n");
		exit (-1);
	}
	// Make sure pointer is on 4K boundary
	if ((unsigned long)gpio_mem % PAGE_SIZE)
	  gpio_mem += PAGE_SIZE - ((unsigned long)gpio_mem % PAGE_SIZE);
	// Now map it
	gpio_map = (char *)mmap
	(
		(caddr_t)gpio_mem,
		BLOCK_SIZE,
		PROT_READ|PROT_WRITE,
		MAP_SHARED|MAP_FIXED,
		mem_fd,
		GPIO_BASE
	);
	if ((long)gpio_map < 0)
		{
		printf("mmap error %d\n", (int)gpio_map);
		exit (-1);
		}
	// Always use volatile pointer!
	gpio = (volatile unsigned *)gpio_map;
	// Now define the outputs:
	// must use INP_GPIO before we can use OUT_GPIO
	INP_GPIO(7);   OUT_GPIO(7);  // Group 1
	INP_GPIO(8);   OUT_GPIO(8);
	INP_GPIO(9);   OUT_GPIO(9);
	INP_GPIO(10);  OUT_GPIO(10); // Group 4
	INP_GPIO(21);  OUT_GPIO(21); // Key 1
	INP_GPIO(22);  OUT_GPIO(22);
	INP_GPIO(23);  OUT_GPIO(23);
	INP_GPIO(24);  OUT_GPIO(24); // Key 4
	INP_GPIO(17);  OUT_GPIO(17); // On
	INP_GPIO(18);  OUT_GPIO(18); // Off
} // Setup_GPIO


int main(int argc, char **argv)
{
  Setup_GPIO(); // Set up gpi pointer for direct register access
  printf("%s\n", VersionString);
  ReadFile();    // once at startup
  SetupTCP();
  printf ("%s\n\n", "Halted....");
  return 0;
} // main

